local lib = LibStub:NewLibrary("LibConfig-1.0", 4)
if not lib then return end
local reg, dlg = LibStub("AceConfigRegistry-3.0"), LibStub("AceConfigDialog-3.0")
lib.data = lib.data or {}

function lib:CreateOptions(name, title, args, handler, slash, dbName, defaults, defaultName, wnd)
	local e = "Usage: CreateOptions(\"name\", [\"title\"] args, handler, \"slash\", dbName, defaults, defaultName, wnd): "
	assert(self ~= lib, e.."do not use LibConfig:CreateOptions(), use your own 'self'")
	if type(title) == "table" then
		title, args, handler, slash, dbName, defaults, defaultName, wnd =
			name, title, args, handler, slash, dbName, defaults, defaultName
	end
	assert(type(args) == "table", e.."'args' - expected table")
	local o, f, db = {
		name = title,
		type = "group",
		handler = self,
		childGroups = "tab",
		inline = true,
		get = "GetOption",
		set = "SetOption",
		args = args
	}
	reg:RegisterOptionsTable(name, o)
	if wnd then f, wnd = wnd, true
	else f = dlg:AddToBlizOptions(name, title)
	end
	lib.data[self].frame[name] = {frame = f, wnd = wnd}
	if handler then lib.data[self].handler = handler end
	if slash then
		assert(type(slash) == "string", e.."slash - expected string")
		LibStub("AceConsole-3.0").RegisterChatCommand(self, slash, function(c)
			if not self.ChatCommand or self:ChatCommand(c) then
				if c == "" then
					if wnd then f:Show()
					else InterfaceOptionsFrame_OpenToCategory(f)
					end
				else LibStub("AceConfigCmd-3.0"):HandleCommand(slash, name, c)
				end
			end
		end)
	end
	if dbName then
		local function OnProfileChanged(event, db, dbName)
			lib.data[self].profile = db.profile
			if self.OnProfileChanged then self:OnProfileChanged(event, db, dbName) end
			for n, d in pairs(lib.data[self].frame) do
				if d.wnd then
					dlg.Status[n] = {children = {}, status = db.profile.status[n]}
					dlg:Open(n, d.frame)
				end
			end
			self:ApplySettings()
			if self.OnLoaded then self:OnLoaded() end
		end
		local function OnProfileShutdown(event, db)
			if self.OnProfileShutdown then self:OnProfileShutdown(event, db) end
		end
		db = LibStub("AceDB-3.0"):New(dbName, defaults, defaultName or "Default")
		db:RegisterCallback("OnProfileChanged", OnProfileChanged)
		db:RegisterCallback("OnProfileReset", OnProfileChanged)
		db:RegisterCallback("OnProfileShutdown", OnProfileShutdown)
		if wnd then
			lib.data[self].frame[name].status = {
				width = defaults.profile.status[name].width,
				height = defaults.profile.status[name].height}
		else
			o.args.profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(db)
		end
		OnProfileChanged("OnProfileChanged", db, db.keys.profile)
	end
end

function lib:GetOption(path, i)
	local p = lib.data[self].profile
	for i = 1, #path - 1 do
		if p[path[i]] then p = p[path[i]] end
	end
	p = p[path[#path]]
	if i then return p[i] else return p end	--return i and p[i] or p: wrong!
end

function lib:SetOption(path, i, v)
	local p, h = lib.data[self].profile, lib.data[self].handler
	for i = 1, #path - 1 do
		if p[path[i]] then
			p = p[path[i]]
			h = h and h[path[i]]
		end
	end
	if v == nil then p[path[#path]], v = i, path, nil
	else p[path[#path]][i] = v
	end
	h = h and h[path[#path]]
	if h then
		if type(h) == "table" then h.func(i, v, path)
		else h(i, v, path)
		end
	end
end

function lib:ExecuteHandler(handler, profile, path)
	if type(handler) == "table" then
		for i, v in pairs(handler) do
			tinsert(path, i)
			self:ExecuteHandler(v, profile[i], path)
			tremove(path)
		end
	elseif type(profile) == "table" then
		for i, v in pairs(profile) do
			handler(i, v, path)
		end
	else handler(profile, path)
	end
end

function lib:ApplySettings()
	local d = lib.data[self]
	if d.handler then
		self:ExecuteHandler(d.handler, d.profile, {})
	end
end

function lib:NotifyChange(name)
	reg:NotifyChange(name)
end

local mixin = {"CreateOptions", "GetOption", "SetOption",
	"ExecuteHandler", "ApplySettings", "NotifyChange"}
function lib:Embed(target)
	for i, v in pairs(mixin) do
		if target[v] ~= lib[v] then target[v] = lib[v] end
	end
	if not lib.data[target] then lib.data[target] = {frame = {}} end
end

--Upgrade old lib
for a in pairs(lib.data) do
	lib:Embed(a)
end
